#include <arpa/inet.h>  /* for inet_ntoa   */
#include <cerrno>       /* for EADDRINUSE  */
#include <csignal>
#include <cstdio>
#include <fcntl.h>      /* for fcntl       */
#include <netdb.h>      /* for gethostbyaddr */
#include <sys/socket.h> /* for socket()    */
#include <sys/stat.h>   /* for O_NONBLOCK  */
#include <unistd.h>
#include <vector>

#include "config.hh"
#include "logfun.h"
#include "radiofun.hh"

httpclient::httpclient(int plug) : AddrLen(sizeof(They)), errorcount(0)
{
    fd = accept(plug, (struct sockaddr *)&They, &AddrLen);
}

void httpclient::MakeNonBlocking()
{
    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
}

void httpclient::PrintHostname() const
{
    char *s = inet_ntoa(They.sin_addr);
    fprintf(stderr, "%s", s); fflush(stderr);
    
    if(Configuration.GetBool("dnslookups"))
    {
        struct hostent *h = gethostbyaddr((const char *)(&They.sin_addr), AddrLen, They.sin_family);
        if(h)
        {
            fprintf(stderr, " (%s)", h->h_name);
            fflush(stderr);
        } 
    }
}

void httpclient::Exec()
{
    pid = fork();
    if(pid)
    {
        if(fd >= 0)
        {
            DumpTime();
            fprintf(stderr, "Got connection from ");
            PrintHostname();
            #if 0
            if(Verbose)
                fprintf(stderr, " - pid %d will handle it", pid);
            #endif
            fprintf(stderr, "\n");
            fflush(stderr);
        }
        return;
    }
    dup2(fd, 0);
    dup2(fd, 1);
    close(fd);
    
    vector<const char *> argv;
    argv.push_back("radio_http");
    if(Server.IsFull())
        argv.push_back("-f");

    argv.push_back(NULL);        
    execv("./radio_http", const_cast<char **>(&argv[0]));
    _exit(-1);
}

bool httpclient::GotClient(int status)
{
    pid = -1;
    if(status == 55)
    {
        MakeNonBlocking();
        return true;
    }
    else
    {
        close(fd);
        return false;
    }
}

class mp3targets : public vector<class httpclient>
{
};

mp3streamer::mp3streamer() : fd(-1)
{
    targets = new mp3targets;
}
mp3streamer::~mp3streamer()
{
    delete targets;
}

void mp3streamer::BeginStreaming()
{
    Configuration.Rehash();
    
    int pip[2];
    pipe(pip);
    pid = fork();
    if(pid)
    {
        DumpTime();
        fprintf(stderr, "Starting encoder.\n");
        close(pip[1]);
        fd = pip[0];
        return;
    }
    dup2(open("/dev/null", O_RDONLY), 0);
    dup2(pip[1], 1);
    close(pip[0]);
    close(pip[1]);
    if(Configuration.GetBool("use_ssh"))
        execl("./radio_ssh", "radio_ssh", NULL);
    else
        execl("./radio_mp3", "radio_mp3", NULL);
    _exit(-1);
}
void mp3streamer::EndStreaming()
{
    DumpTime();
    fprintf(stderr, "No more listeners, halting encoder.\n");
    close(fd);
    kill(pid, SIGTERM);
    fd = -1;
}
void mp3streamer::addtarget(const httpclient &client)
{
    DumpTime();
    client.PrintHostname();
    fprintf(stderr, " begins listening.\n");
    
    if(fd < 0)BeginStreaming();
    targets->push_back(client);
}

bool mp3streamer::IsFull() const
{
    unsigned kbrate = Configuration.GetInteger("kbrate");
    unsigned traffic = targets->size() * kbrate;
    return traffic + kbrate > Configuration.GetInteger("maxtraffic");
}

void mp3streamer::forward()
{
    char Buf[8192];
    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    int c = read(fd, Buf, sizeof Buf);
    if(c <= 0)
    {
        if(c < 0 && errno == EAGAIN)return;
        
        DumpTime();
        fprintf(stderr, "Encoder succofated. Disconnecting all listeners.\n");
        
        /* Stream ended */
        for(unsigned a=0; a<targets->size(); ++a)
        {
            DumpTime();
            fprintf(stderr, "- ");
            (*targets)[a].PrintHostname();
            fprintf(stderr, " drops now.\n");
            close((*targets)[a].fd);
        }
        targets->clear();
        EndStreaming();
    }
    else
    {
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
        for(unsigned a=0; a<targets->size(); )
        {
            int olc = fcntl(fd, F_GETFL);
            fcntl((*targets)[a].fd, F_SETFL, olc & ~O_NONBLOCK);
            int i = write((*targets)[a].fd, Buf, c);
            fcntl((*targets)[a].fd, F_SETFL, olc);
            
            if(i < 0 && errno == EPIPE)
            {
                DumpTime();
                fprintf(stderr, "Client ");
                (*targets)[a].PrintHostname();
                fprintf(stderr, " signed off, dropping.\n");
                goto Drop;
            }
            if(i < c)
            {
                if(i >= 0)
                {
                    DumpTime();
                    fprintf(stderr, "Could only send %u bytes to ", i);
                    (*targets)[a].PrintHostname();
                    fprintf(stderr, " (tried %u)\n", c);
                }
                /* Max 10 errors */
                
                (*targets)[a].errorcount += 20;
                if((*targets)[a].errorcount > 200)
                {
                    DumpTime();
                    fprintf(stderr, "Too many errors for ");
                    (*targets)[a].PrintHostname();
                    fprintf(stderr, " - dropping.\n");
                Drop:
                    close((*targets)[a].fd);
                    targets->erase(targets->begin() + a);
                    continue;
                }
            }
            else
            {
                /* The error counter resets if no errors for a long time */
                if((*targets)[a].errorcount > 0)
                    --(*targets)[a].errorcount;
            }
            ++a;
        }
        if(!targets->size())
            EndStreaming();
    }
}

mp3streamer Server;
