#include <cerrno>
#include <cstdio>
#include <cstring>     /* for memset */
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h> /* for inet_ntoa */
#include <unistd.h>
#include <fcntl.h>

#include <string>

#include "mixer.h"
#include "argh.hh"

static string A0;

#define VERSIONSTR \
    "%s v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "This program is under GPL. wavegenerator-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n", A0.c_str()

string Host = "localhost";
string StreamForm;

class Handle : public arghandler
{
    int blur;
    string filter;
private:
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_V(const string &)
    {
        printf(VERSIONSTR);
        exit(0);
    }
    string opt_b(const string &s)
    {
        unsigned a=0;
        for(blur=0; a<s.size() && isdigit(s[a]); )blur = blur*10 + s[a++] - '0';
        blur = 1+blur;
        return s.substr(a);
    }
    string opt_s(const string &s)
    {
        unsigned a=0;
        for(blur=0; a<s.size() && isdigit(s[a]); )blur = blur*10 + s[a++] - '0';
        blur = 1 - blur;
        return s.substr(a);
    }
    string opt_f(const string &s) { filter     = s; return ""; }
    string opt_H(const string &s) { Host       = s; return ""; }
    string opt_F(const string &s) { StreamForm = s; return ""; }
public:
    bool Help;
public:
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        blur(1), Help(false)
    {
        add("-h", "--help",     "", &(Handle::opt_h));
        add("-V", "--version",  "", &(Handle::opt_V));
        add("-f", "--filter",   "", &(Handle::opt_f));
        add("-F", "--format",   "", &(Handle::opt_F));
        add("-b", "--blur",     "", &(Handle::opt_b));
        add("-H", "--host",     "", &(Handle::opt_H));
        add("-s", "--sharpen",  "", &(Handle::opt_s));
    }
    virtual void defarg(const string &)
    {
        //argerror(s, true);
    }
    const string GetFilter() const
    {
        char Buf[64] = "";
        if(blur < 1)
            sprintf(Buf, "s%d", 1-blur);
        else if(blur > 1)
            sprintf(Buf, "b%d", blur-1);
        string p = Buf;
        if(filter.size())p += "f" + filter;
        if(!p.size())return "";
        return "-"+p;
    }
    void Usage(const string &specs, const string &ops)
    {
        fprintf(stderr,
            "Usage: %s [<options> ...]\n"
            "\n"
            "%s.\n"
            "\n",
            A0.c_str(),
            (A0=="mixmon" ? "Writes stdout" : "Reads stdin"));
        fprintf(stderr,
            "Stream format strings (examples):\n"
            "%s\n"
            "\n"
            "Options:\n"
            "  -H, --host    Specify host where mixer is running\n"
            "  -F, --format  Stream format string, see above\n"
            "%s\n",
            specs.c_str(),
            ops.c_str());
    }
};

static string GetError(int fd)
{
    char c = 0;
    read(fd, &c, 1);
    if(c != '!') /* Get error message if any */
    {
        string err;
        do {
            err += c;
        } while(read(fd, &c, 1) == 1);
        return err;
    }
    return "";
}

int main(int argc, const char *const *argv)
{
    const char *a0 = argv[0];
    for(const char *q; (q = strchr(a0, '/')); )a0 = q+1;
    A0 = a0;
    
    char type = A0 != "mixmon"
                   ? 'S'  /* stream */
                   : 'M'; /* monitor */
                   
    Handle parameters("", argc, argv);
    parameters.parse();
    
    int fd;
    
    if(parameters.Help) /* just want the help text */
    {
        int pip[2];
        pipe(pip);
        int pid = fork();
        if(!pid)
        {
            close(pip[0]);
            dup2(pip[1], 1);
            dup2(pip[1], 2);
            close(pip[1]);
            execlp("resample", "resample", "--help", NULL);
            fprintf(stderr, VERSIONSTR);
            fprintf(stderr,
            	"\n"
            	"resample not found.\n"
            	"It is required. It is normally included in the\n"
            	"same package where this program came from.\n"
            );
            _exit(-1);
        }
        close(pip[1]);
        fd = pip[0];
    }
    else
    {
        struct sockaddr_in sin;
        sin.sin_port = htons(MIXER_PORT);
        struct hostent *hent = gethostbyname(Host.c_str());
        if(!hent)
        {
            fprintf(stderr, "Can't resolve hostname \"%s\"\n", Host.c_str());
            return -1;
        }
        memcpy(&sin.sin_addr, hent->h_addr_list[0], sizeof(struct in_addr));
        fd = socket(sin.sin_family = AF_INET, SOCK_STREAM, 0);
        if(connect(fd, (struct sockaddr *)&sin, sizeof(struct sockaddr)) < 0)
        {
            close(fd);
            fprintf(stderr, "Unable to connect to \"%s\"", Host.c_str());
            perror("");
            return errno;
        }
        
        static const char lf = '\n';
        write(fd, &type, 1);
        string FilterParm = parameters.GetFilter();
        write(fd, StreamForm.data(), StreamForm.size()); write(fd, &lf, 1);
        write(fd, FilterParm.data(), FilterParm.size()); write(fd, &lf, 1);
    }
    
    string error = GetError(fd);
    if(error.size())
    {
        unsigned spec = error.find("spec:");
        unsigned opts = error.find("Options:");
        if(spec == error.npos
        || opts == error.npos)
        {
            fprintf(stderr, "%s", error.c_str());
            return -1;
        }
        string specs = error.substr(spec+6,opts-spec-8);
        string ops = error.substr(opts+9, error.size()-opts-10);
        
        parameters.Usage(specs, ops);
        return -1;
    }
    
    int src  = type=='S' ? 0 : fd;
    int dest = type=='M' ? 1 : fd;
    
    for(;;)
    {
        char Buf[4096];
        int c = read(src, Buf, sizeof Buf);
        if(c <= 0)break;
        
        /* We get SIGPIPE when this fails, so no worth checking */
        write(dest, Buf, c);
    }
        
    return 0;
}
